@@ -3,8 +3,18 @@ import re
 import sys
 
 from sly import Parser
+from sly.yacc import SlyLogger
 from json5.tokenizer import JSONLexer, tokenize
 from json5.model import *
+from json5.utils import JSON5DecodeError
+
+
+class QuietSlyLogger(SlyLogger):
+    def warning(self, *args, **kwargs):
+        return
+    debug = warning
+    info = warning
+
 
 ESCAPE_SEQUENCES = {
     'b': '\u0008',
@@ -31,52 +41,62 @@ def replace_escape_literals(matchobj):
 class JSONParser(Parser):
     # debugfile = 'parser.out'
     tokens = JSONLexer.tokens
+    log = QuietSlyLogger(sys.stderr)
+
+    def __init__(self, *args, **kwargs):
+        super().__init__(*args, **kwargs)
+        self.errors = []
 
 
-    @_('value')
+    @_('{ wsc } value { wsc }')
     def text(self, p):
-        return JSONText(value=p[0])
+        node = JSONText(value=p[1])
+        for wsc in p.wsc0:
+            node.wsc_before.append(wsc)
+        for wsc in p.wsc1:
+            node.wsc_after.append(wsc)
+        return node
 
-
-    @_('key COLON value')
+    @_('key { wsc } COLON { wsc } value { wsc }')
     def first_key_value_pair(self, p):
+        key = p[0]
+        for wsc in p.wsc0:
+            key.wsc_after.append(wsc)
+        value = p[4]
+        for wsc in p.wsc1:
+            value.wsc_before.append(wsc)
+        for wsc in p.wsc2:
+            value.wsc_after.append(wsc)
         return KeyValuePair(key=p.key, value=p.value)
 
-    @_('COMMA { whitespace_andor_comment } [ key COLON value ]')
+
+    @_('COMMA { wsc } [ first_key_value_pair ]')
     def subsequent_key_value_pair(self, p):
-        if p.key and p.value:
-            for wsc in p.whitespace_andor_comment:
-                p.key.wsc_before.append(wsc)
-            return KeyValuePair(key=p.key, value=p.value)
+        if p.first_key_value_pair:
+            node = p.first_key_value_pair
+            for wsc in p.wsc:
+                node.key.wsc_before.append(wsc)
         else:
             node = TrailingComma()
-            for wsc in p.whitespace_andor_comment:
+            for wsc in p.wsc:
                 node.wsc_after.append(wsc)
-            return node
+        return node
 
 
     @_('WHITESPACE',
        'comment')
-    def whitespace_andor_comment(self, p):
+    def wsc(self, p):
         return p[0]
 
-    @_('[ WHITESPACE ] BLOCK_COMMENT [ WHITESPACE ]')
+    @_('BLOCK_COMMENT')
     def comment(self, p):
-        node = BlockComment(p[1])
-        if p.WHITESPACE0:
-            node.wsc_before.append(p.WHITESPACE0)
-        if p.WHITESPACE1:
-            node.wsc_after.append(p.WHITESPACE1)
-        return node
+        return BlockComment(p[0])
 
-    @_('[ WHITESPACE ] LINE_COMMENT [ WHITESPACE ]')
+
+    @_('LINE_COMMENT')
     def comment(self, p):
-        node = LineComment(p[1])
-        if p.WHITESPACE0:
-            node.wsc_before.append(p.WHITESPACE0)
-        if p.WHITESPACE1:
-            node.wsc_after.append(p.WHITESPACE1)
-        return node
+        return LineComment(p[0])
+
 
 
     @_('first_key_value_pair { subsequent_key_value_pair }')
@@ -90,28 +110,33 @@ class JSONParser(Parser):
 
 
 
-    @_('LBRACE [ key_value_pairs ] RBRACE')
+    @_('LBRACE { wsc } [ key_value_pairs ] RBRACE')
     def json_object(self, p):
         if not p.key_value_pairs:
-            return JSONObject()
-        kvps, trailing_comma = p.key_value_pairs
-        return JSONObject(*kvps, trailing_comma=trailing_comma)
+            node = JSONObject(leading_wsc=list(p.wsc or []))
+        else:
+            kvps, trailing_comma = p.key_value_pairs
+            node = JSONObject(*kvps, trailing_comma=trailing_comma, leading_wsc=list(p.wsc or []))
 
-    @_('value')
+        return node
+
+    @_('value { wsc }')
     def first_array_value(self, p):
-        return p[0]
+        node = p[0]
+        for wsc in p.wsc:
+            node.wsc_after.append(wsc)
+        return node
 
-    @_('COMMA { whitespace_andor_comment } [ value ]')
+    @_('COMMA { wsc } [ first_array_value ]')
     def subsequent_array_value(self, p):
-        if p.value:
-            node = p.value
-            for wsc in p.whitespace_andor_comment:
-                node.wsc_before.append(wsc)  # Wouldn't the value already take care of this? reduce...reduce....
+        if p.first_array_value:
+            node = p.first_array_value
+            for wsc in p.wsc:
+                node.wsc_before.append(wsc)
         else:
             node = TrailingComma()
-            for wsc in p.whitespace_andor_comment:
+            for wsc in p.wsc:
                 node.wsc_after.append(wsc)
-
         return node
 
     @_('first_array_value { subsequent_array_value }')
@@ -124,27 +149,27 @@ class JSONParser(Parser):
         return ret, None
 
 
-    @_('LBRACKET [ array_values ] RBRACKET')
+    @_('LBRACKET { wsc } [ array_values ] RBRACKET')
     def json_array(self, p):
         if not p.array_values:
-            return JSONArray()
-        values, trailing_comma = p.array_values
-        return JSONArray(*values, trailing_comma=trailing_comma)
+            node = JSONArray()
+        else:
+            values, trailing_comma = p.array_values
+            node = JSONArray(*values, trailing_comma=trailing_comma)
 
+        for wsc in p.wsc:
+            node.leading_wsc.append(wsc)
 
+        return node
 
     @_('NAME')
     def identifier(self, p):
         return Identifier(name=p[0])
 
-    @_('{ whitespace_andor_comment } identifier { whitespace_andor_comment }',
-       '{ whitespace_andor_comment } string { whitespace_andor_comment }')
+    @_('identifier',
+       'string')
     def key(self, p):
-        node = p[1]
-        for wsc in p.whitespace_andor_comment0:
-            node.wsc_before.append(wsc)
-        for wsc in p.whitespace_andor_comment1:
-            node.wsc_after.append(wsc)
+        node = p[0]
         return node
 
     @_('INTEGER')
@@ -163,24 +188,16 @@ class JSONParser(Parser):
     def number(self, p):
         return NaN()
 
-    @_('{ whitespace_andor_comment } MINUS number { whitespace_andor_comment }')
+    @_('MINUS number')
     def value(self, p):
         if isinstance(p.number, Infinity):
             p.number.negative = True
         node = UnaryOp(op='-', value=p.number)
-        for wsc in p.whitespace_andor_comment0:
-            node.wsc_before.append(wsc)
-        for wsc in p.whitespace_andor_comment1:
-            node.wsc_after.append(wsc)
         return node
 
-    @_('{ whitespace_andor_comment } PLUS number { whitespace_andor_comment }')
+    @_('PLUS number')
     def value(self, p):
         node = UnaryOp(op='+', value=p.number)
-        for wsc in p.whitespace_andor_comment0:
-            node.wsc_before.append(wsc)
-        for wsc in p.whitespace_andor_comment1:
-            node.wsc_after.append(wsc)
         return node
 
     @_('INTEGER EXPONENT',
@@ -197,6 +214,9 @@ class JSONParser(Parser):
     def double_quoted_string(self, p):
         raw_value = p[0]
         contents = raw_value[1:-1]
+        if re.search(r'(?<!\\)([\u000D\u2028\u2029]|(?<!\r)\n)', contents):
+            errmsg = f"Illegal line terminator without continuation"
+            self.errors.append(JSON5DecodeError(errmsg, p._slice[0]))
         contents = re.sub(r'\\(\r\n|[\u000A\u000D\u2028\u2029])', '', contents)
         contents = re.sub(r'\\(.)', replace_escape_literals, contents)
         return DoubleQuotedString(contents, raw_value=raw_value)
@@ -205,6 +225,9 @@ class JSONParser(Parser):
     def single_quoted_string(self, p):
         raw_value = p[0]
         contents = raw_value[1:-1]
+        if re.search(r'(?<!\\)([\u000D\u2028\u2029]|(?<!\r)\n)', contents):
+            errmsg = f"Illegal line terminator without continuation"
+            self.errors.append(JSON5DecodeError(errmsg, p._slice[0]))
         contents = re.sub(r'\\(\r\n|[\u000A\u000D\u2028\u2029])', '', contents)
         contents = re.sub(r'\\(.)', replace_escape_literals, contents)
         return SingleQuotedString(contents, raw_value=raw_value)
@@ -226,18 +249,14 @@ class JSONParser(Parser):
     def null(self, p):
         return NullLiteral()
 
-    @_('{ whitespace_andor_comment } string { whitespace_andor_comment }',
-       '{ whitespace_andor_comment } json_object { whitespace_andor_comment }',
-       '{ whitespace_andor_comment } json_array { whitespace_andor_comment }',
-       '{ whitespace_andor_comment } boolean { whitespace_andor_comment }',
-       '{ whitespace_andor_comment } null { whitespace_andor_comment }',
-       '{ whitespace_andor_comment } number { whitespace_andor_comment }',)
+    @_('string',
+       'json_object',
+       'json_array',
+       'boolean',
+       'null',
+       'number',)
     def value(self, p):
-        node = p[1]
-        for wsc in p.whitespace_andor_comment0:
-            node.wsc_before.append(wsc)
-        for wsc in p.whitespace_andor_comment1:
-            node.wsc_after.append(wsc)
+        node = p[0]
         return node
 
     @_('BREAK',
@@ -267,7 +286,34 @@ class JSONParser(Parser):
     'IN',
     'TRY',)
     def identifier(self, p):
-        raise SyntaxError(f"{p[0]} is a reserved keyword and may not be used")
+        err = JSON5DecodeError(f"Illegal name ({p[0]}) is a reserved keyword and may not be used", p._slice[0])
+        self.errors.append(err)
+        return Identifier(name=p[0])
+
+    def error(self, token):
+        if token:
+            self.errors.append(JSON5DecodeError('Syntax Error', token))
+        else:
+            self.errors.append(JSON5DecodeError('Expecting value. Received unexpected EOF', None))
+
+    def parse(self, *args, **kwargs):
+        model = super().parse(*args, **kwargs)
+        if self.errors:
+            if len(self.errors) > 1:
+                primary_error = self.errors[0]
+                additional_errors = '\n\t'.join(err.args[0] for err in self.errors[1:])
+                msg = "There were multiple errors parsing the JSON5 document.\n" \
+                      "The primary error was: \n\t{}\n" \
+                      "Additionally, the following errors were also detected:\n\t {}"
+                msg = msg.format(primary_error.args[0], additional_errors)
+                err = JSON5DecodeError(msg, None)
+                err.lineno = primary_error.lineno
+                err.token = primary_error.token
+                err.index = primary_error.index
+                raise err
+            else:
+                raise self.errors[0]
+        return model
 
 
 def parse_tokens(raw_tokens):
@@ -278,6 +324,4 @@ def parse_tokens(raw_tokens):
 def parse_source(text):
     tokens = tokenize(text)
     model = parse_tokens(tokens)
-    if model is None:
-        raise SyntaxError('Was expecting a JSON value, got none. This was probably due to a syntax error while parsing')
     return model
